import 'package:test/test.dart';
import 'package:got_it/got_it.dart';

/// Dummy object created to only equal itself
abstract class Unique {}

/// Of different types to notice that the item got replaced
class UniqueA implements Unique {}

class UniqueB implements Unique {}

void main() {
  group("Creation", () {
    group("of an instance", () {
      test("successfully returns an instance of GotIt", () {
        expect(GotIt(), TypeMatcher<GotIt>());
      });
      test("returns different instances", () {
        final firstInvocation = GotIt();
        final secondInvocation = GotIt();
        expect(firstInvocation == secondInvocation, isFalse);
      });
    });

    group("of a singleton", () {
      test("successfully returns an instance of GotIt", () {
        expect(GotIt.singleton(), TypeMatcher<GotIt>());
      });
      test("always returns the same instance", () {
        final firstInvocation = GotIt.singleton();
        final secondInvocation = GotIt.singleton();
        expect(firstInvocation == secondInvocation, isTrue);
      });
    });
  });

  group("Registering", () {
    group("a factory", () {
      group("once", () {
        test("returns unique items each invocation", () {
          final gotIt = GotIt();

          gotIt.registerFactory<Unique>(() => UniqueA());

          final firstInvocation = gotIt<Unique>();
          final secondInvocation = gotIt<Unique>();

          expect(firstInvocation == secondInvocation, isFalse);
        });
      });
      group("twice", () {
        test("when allowReassignment is false throws", () {
          final gotIt = GotIt();

          gotIt.registerFactory<Unique>(
            () => UniqueA(),
            allowReassignment: false,
          );

          expect(() => gotIt.registerFactory<Unique>(() => UniqueB()),
              throwsA(TypeMatcher<AssertionError>()));
        });
        test("returns a different object", () {
          final gotIt = GotIt();

          gotIt.registerFactory<Unique>(
            () => UniqueA(),
            allowReassignment: true,
          );
          final firstInvocation = gotIt<Unique>();

          gotIt.registerFactory<Unique>(() => UniqueB());
          final secondInvocation = gotIt<Unique>();

          expect(firstInvocation, TypeMatcher<UniqueA>());
          expect(secondInvocation, TypeMatcher<UniqueB>());
        });
      });
    });
    group("a lazy singleton", () {
      group("once", () {
        test("returns the same items each time", () {
          final gotIt = GotIt();

          gotIt.registerLazySingleton<Unique>(() => UniqueA());

          final firstInvocation = gotIt<Unique>();
          final secondInvocation = gotIt<Unique>();

          expect(firstInvocation == secondInvocation, isTrue);
        });
      });
      group("twice", () {
        test("when allowReassignment is false throws", () {
          final gotIt = GotIt();

          gotIt.registerLazySingleton<Unique>(
            () => UniqueA(),
            allowReassignment: false,
          );

          expect(() => gotIt.registerLazySingleton<Unique>(() => UniqueB()),
              throwsA(TypeMatcher<AssertionError>()));
        });
        test("returns a different object", () {
          final gotIt = GotIt();

          gotIt.registerLazySingleton<Unique>(
            () => UniqueA(),
            allowReassignment: true,
          );
          final firstInvocation = gotIt<Unique>();

          gotIt.registerLazySingleton<Unique>(() => UniqueB());
          final secondInvocation = gotIt<Unique>();

          expect(firstInvocation, TypeMatcher<UniqueA>());
          expect(secondInvocation, TypeMatcher<UniqueB>());
        });
      });
    });
    group("a singleton", () {
      group("once", () {
        test("returns the same item it was given each time", () {
          final gotIt = GotIt();

          final unique = UniqueA();

          gotIt.registerSingleton<Unique>(unique);

          final firstInvocation = gotIt<Unique>();
          final secondInvocation = gotIt<Unique>();

          expect(firstInvocation == secondInvocation, isTrue);
          expect(firstInvocation == unique, isTrue);
        });
      });
      group("twice", () {
        test("when allowReassignment is false throws", () {
          final gotIt = GotIt();

          gotIt.registerSingleton<Unique>(
            UniqueA(),
            allowReassignment: false,
          );

          expect(() => gotIt.registerSingleton<Unique>(UniqueB()),
              throwsA(TypeMatcher<AssertionError>()));
        });
        test("returns a different object", () {
          final gotIt = GotIt();

          gotIt.registerSingleton<Unique>(
            UniqueA(),
            allowReassignment: true,
          );
          final firstInvocation = gotIt<Unique>();

          gotIt.registerSingleton<Unique>(UniqueB());
          final secondInvocation = gotIt<Unique>();

          expect(firstInvocation, TypeMatcher<UniqueA>());
          expect(secondInvocation, TypeMatcher<UniqueB>());
        });
      });
    });
  });

  test("Getting a type with no associated registrations throws", () {
    final gotIt = GotIt();
    expect(() => gotIt<Unique>(), throwsA(TypeMatcher<UnregisteredError>()));
  });

  test("Resetting GotIt empties any existing registration", () {
    final gotIt = GotIt();
    gotIt.registerSingleton<Unique>(UniqueA());
    expect(() => gotIt<Unique>(), returnsNormally);
    gotIt.reset();
    expect(() => gotIt<Unique>(), throwsA(TypeMatcher<UnregisteredError>()));
  });

  test(
      "Using the singleton in two different functions"
      "returns the same registered object", () {
    final unique = UniqueA();
    void methodOne() {
      final gotIt = GotIt.singleton();
      gotIt.registerSingleton<Unique>(unique);
    }

    Unique methodTwo() {
      final gotIt = GotIt.singleton();
      return gotIt<Unique>();
    }

    methodOne();

    expect(methodTwo() == unique, isTrue);
  });
}
